# -*- coding: utf-8 -*-
""" SUBCOMMAND:  ``combine --target=<DIR> <SRC1> [ <SRC-n> ]``

Usage example:

.. code-block:: sh

    cd MY_APP
    ksconf combine default.d/* --target=default

"""
from __future__ import absolute_import, unicode_literals

import os
from io import open

from ksconf.combine import LayerCombiner, LayerCombinerException
from ksconf.commands import KsconfCmd, dedent
from ksconf.consts import (EXIT_CODE_BAD_ARGS, EXIT_CODE_COMBINE_MARKER_MISSING,
                           EXIT_CODE_MISSING_ARG, EXIT_CODE_NO_SUCH_FILE)
from ksconf.filter import create_filtered_list
from ksconf.util.completers import DirectoriesCompleter
from ksconf.util.file import expand_glob_list, relwalk, splglob_simple

CONTROLLED_DIR_MARKER = ".ksconf_controlled"


class LayerCombinerExceptionCode(LayerCombinerException):
    def __init__(self, msg, return_code=None):
        super().__init__(msg)
        self.return_code = return_code


class RepeatableCombiner(LayerCombiner):
    """
    Re-runable combiner class.  Beyond the reusable layer combining functionality,
    this class enables the use of a marker file for added safety.  Removed files
    will cleanup.
    """

    def __init__(self, *args,
                 disable_marker: bool = False,
                 disable_cleanup: bool = False,
                 keep_existing: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_extra_files: set = None
        self.disable_marker = disable_marker
        self.disable_cleanup = disable_cleanup
        self.keep_existing = keep_existing

    def prepare_target_dir(self, target):
        """
        Handle marker file and ensure that target directory gets created safely.
        """
        marker_file = os.path.join(target, CONTROLLED_DIR_MARKER)
        if os.path.isdir(target):
            if not self.disable_marker and not os.path.isfile(marker_file):
                self.stderr.write("Target directory already exists, but it appears to have been "
                                  "created by some other means.  Marker file missing.\n")
                raise LayerCombinerExceptionCode("Target directory exists without marker file",
                                                 EXIT_CODE_COMBINE_MARKER_MISSING)

        elif self.dry_run:
            self.stderr.write("Skipping creating destination directory {target} (dry-run)\n")
        else:
            try:
                os.mkdir(target)
            except OSError as e:
                self.stderr.write(f"Unable to create destination directory {target}.  {e}\n")
                raise LayerCombinerExceptionCode(f"Unable to create destination directory {target}",
                                                 EXIT_CODE_NO_SUCH_FILE)
            self.stderr.write(f"Created destination directory {target}\n")
            if not self.disable_marker:
                with open(marker_file, "w") as f:
                    f.write("This directory is managed by KSCONF.  Don't touch\n")

    def pre_combine_inventory(self, target, src_files):
        """
        Find a set of files that exist in the target folder, but in NO source folder (for cleanup)
        """
        config = self.config

        self.stderr.write(f"Layers detected:  {self.layer_names_all}\n")
        if self.layer_names_all != self.layer_names_used:
            self.stderr.write(f"Layers after filter: {self.layer_names_used}\n")

        # Convert src_files to a set to speed up
        src_files = set(src_files)
        self.target_extra_files = set()
        for (root, dirs, files) in relwalk(target, followlinks=config.follow_symlink):
            for fn in files:
                tgt_file = os.path.join(root, fn)
                if tgt_file not in src_files:
                    if fn == CONTROLLED_DIR_MARKER or config.block_files.search(fn):
                        continue  # pragma: no cover (peephole optimization)
                    self.target_extra_files.add(tgt_file)
        return src_files

    def post_combine(self, target):
        """
        Handle cleanup of extra files
        """
        target_extra_files = self.target_extra_files
        if target_extra_files:
            if self.disable_cleanup:
                self.stderr.write("Cleanup operations disabled by user.\n")
            else:
                self.stderr.write("Found extra files not part of source tree(s):  "
                                  f"{len(target_extra_files)} files.\n")

            keep_existing = create_filtered_list("splunk", default=False)
            # splglob_simple:  Either full paths, or simple file-only match
            keep_existing.feedall(self.keep_existing, filter=splglob_simple)
            for dest_fn in target_extra_files:
                if keep_existing.match_path(dest_fn):
                    self.stderr.write(f"Keep existing file {dest_fn}\n")
                elif self.disable_cleanup:
                    self.stderr.write(f"Skip cleanup of unwanted file {dest_fn}\n")
                else:
                    self.stderr.write(f"Remove unwanted file {dest_fn}\n")
                    os.unlink(os.path.join(target, dest_fn))


class CombineCmd(KsconfCmd):
    help = dedent("""\
    Combine configuration files across multiple source directories into a single
    destination directory.  This allows for an arbitrary number of Splunk
    configuration layers to coexist within a single app.  Useful in both ongoing
    merge and one-time ad-hoc use.
    """)
    description = dedent("""\
    Merge .conf settings from multiple source directories into a combined target
    directory.  Configuration files can be stored in a ``/etc/*.d`` like directory
    structure and consolidated back into a single 'default' directory.

    This command supports both one-time operations and recurring merge jobs.  For
    example, this command can be used to combine all users' knowledge objects (stored
    in 'etc/users') after a server migration, or to merge a single user's settings
    after their account has been renamed.  Recurring operations assume some type
    of external scheduler is being used.  A best-effort is made to only write to
    target files as needed.

    The 'combine' command takes your logical layers of configs (upstream, corporate,
    Splunk admin fixes, and power user knowledge objects, ...) expressed as
    individual folders and merges them all back into the single ``default`` folder
    that Splunk reads from.  One way to keep the 'default' folder up-to-date is
    using client-side git hooks.

    No directory layout is mandatory, but taking advantages of the native-support
    for 'dir.d' layout works well for many uses cases.  This idea is borrowed from
    the Unix System V concept where many services natively read their config files
    from ``/etc/*.d`` directories.

    Version notes:  dir.d was added in ksconf 0.8.  Starting in 1.0 the default will
    switch to 'dir.d', so if you need the old behavior be sure to update your scripts.
    """)
    format = "manual"
    maturity = "beta"

    def register_args(self, parser):

        def wb_type(action):
            def f(pattern):
                return action, pattern
            return f

        parser.add_argument("source", nargs="+", help=dedent("""
            The source directory where configuration files will be merged from.
            When multiple source directories are provided, start with the most general and end
            with the specific; later sources will override values from the earlier ones.
            Supports wildcards so a typical Unix ``conf.d/##-NAME`` directory structure works well.""")
                            ).completer = DirectoriesCompleter()
        parser.add_argument("--target", "-t", help=dedent("""
            Directory where the merged files will be stored.
            Typically either 'default' or 'local'""")
                            ).completer = DirectoriesCompleter()
        parser.add_argument("-m", "--layer-method",
                            choices=["auto", "dir.d", "disable"],
                            default="auto",
                            help="""
            Set the layer type used by SOURCE.

            Use ``dir.d`` if you have directories like ``MyApp/default.d/##-layer-name``, or use
            ``disable`` to manage layers explicitly and avoid any accidental layer detection.
            By default, ``auto`` mode will enable transparent switching between 'dir.d' and 'disable'
            (legacy) behavior.
            """)

        parser.add_argument("-q", "--quiet", action="store_true",
                            help="Make output a bit less noisy.  This may change in the future...")

        parser.add_argument("-I", "--include", action="append", default=[], dest="layer_filter",
                            type=wb_type("include"), metavar="PATTERN",
                            help="Name or pattern of layers to include.")
        parser.add_argument("-E", "--exclude", action="append", default=[], dest="layer_filter",
                            type=wb_type("exclude"), metavar="PATTERN",
                            help="Name or pattern of layers to exclude from the target.")
        parser.add_argument("--dry-run", "-D", default=False, action="store_true", help=dedent("""
            Enable dry-run mode.
            Instead of writing to TARGET, preview changes as a 'diff'.
            If TARGET doesn't exist, then show the merged file."""))
        parser.add_argument("--follow-symlink", "-l", action="store_true", default=False,
                            help="Follow symbolic links pointing to directories.  "
                                 "Symlinks to files are always followed.")
        parser.add_argument("--banner", "-b",
                            default=" **** WARNING: This file is managed by 'ksconf combine', do "
                                    "not edit hand-edit this file! ****",
                            help="A banner or warning comment added to the top of the TARGET file. "
                                 "Used to discourage Splunk admins from editing an auto-generated "
                                 "file.")
        parser.add_argument("-K", "--keep-existing", action="append", default=[],
                            help="Existing file(s) to preserve in the TARGET folder.  "
                            "This argument may be used multiple times.")
        parser.add_argument("--disable-marker", action="store_true", default=False, help=dedent(f"""
            Prevents the creation of or checking for the ``{CONTROLLED_DIR_MARKER}`` marker file safety check.
            This file is typically used indicate that the destination folder is managed by ksconf.
            This option should be reserved for well-controlled batch processing scenarios.
            """))
        parser.add_argument("--disable-cleanup", action="store_true", default=False,
                            help="Disable all file removal operations.  Skip the cleanup phase "
                            "that typically removes files in TARGET that no longer exist in SOURCE")

    def run(self, args):
        combiner = RepeatableCombiner(follow_symlink=args.follow_symlink, banner=args.banner, dry_run=args.dry_run, quiet=args.quiet,
                                      keep_existing=args.keep_existing, disable_cleanup=args.disable_cleanup, disable_marker=args.disable_marker)

        # For now, just copy all settings from 'args' to class instance... needs work
        combiner.stdout = self.stdout
        combiner.stderr = self.stderr

        for (action, pattern) in args.layer_filter:
            combiner.add_layer_filter(action, pattern)

        # Expand any globs in the CLI to individual directories
        args.source = list(expand_glob_list(args.source, do_sort=True))

        if args.layer_method == "auto":
            self.stderr.write(
                "Warning:  Automatically guessing an appropriate directory layer detection.  "
                "Consider using '--layer-method' to avoid this warning.\n")
            if len(args.source) == 1:
                layer_method = "dir.d"
            else:
                layer_method = "disable"
        else:
            layer_method = args.layer_method

        if layer_method == "dir.d":
            self.stderr.write("Using automatic '*.d' directory layer detection.\n")
            if len(args.source) > 1:
                # XXX: Lift this restriction, if possible.  Seems like this *should* be doable. idk
                self.stderr.write("ERROR:  Only one source directory is allowed when running the "
                                  "'dir.d' layer mode.\n")
                return EXIT_CODE_BAD_ARGS

            combiner.set_layer_root(args.source[0])
            layer_root = combiner.layer_root
            for (dir, layers) in layer_root._mount_points.items():
                self.stderr.write(f"Found layer parent folder:  {dir}  "
                                  f"with layers {', '.join(layers)}\n")
        else:
            self.stderr.write("Automatic layer detection is disabled.\n")
            for src in args.source:
                self.stderr.write(f"Reading conf files from directory {src}\n")
            combiner.set_source_dirs(args.source)

        if args.target is None:
            self.stderr.write("Must provide the '--target' directory.\n")
            return EXIT_CODE_MISSING_ARG

        self.stderr.write(f"Combining files into directory {args.target}\n")

        try:
            combiner.combine(args.target)
        except LayerCombinerExceptionCode as e:
            return e.return_code



#!/usr/bin/env python
from __future__ import absolute_import, print_function, unicode_literals

import os
import re
from textwrap import dedent

from setuptools import setup

from ksconf.setup_entrypoints import get_entrypoints_setup

package_name = "kintyre-splunk-conf" if os.getenv("BUILD_OLD_PACKAGE") == "1" else "ksconf"

if package_name == "ksconf":
    package_rename_message = "Earlier version of this project were packaged as " \
        "[kintyre-splunk-conf](https://pypi.org/project/kintyre-splunk-conf) prior to v0.10.0"
else:
    package_rename_message = dedent(f"""
    ## Package rename
    This package has been renamed [ksconf](https://pypi.org/project/ksconf/) starting with v0.10.
    Please switch to this new package using the upgrade steps listed below.

    ## Suggested upgrade steps:
    ```sh
    pip uninstall {package_name}
    pip install ksconf
    ```

    ## Who should keep using {package_name}?
    For the time being, you can still install and upgrade the latest release of
    ksconf using the `{package_name}` package.

    Note that as of the v0.10 release, only Python 3.7 and higher are supported.
    If you need a version of ksconf that works with Python 2.7 & Python 3.6,
    then you should use {package_name} prior to v0.10.  We suggest that everyone
    else should upgrade to the latest `ksconf` package.

    Side note:  Attempts to make {package_name} simply install the new ksconf
    package under the covers resulted in an unusable deployment.  So for now
    we are building two identical packages under different names.  In the next
    minor release you can expect additional onscreen warnings about the package
    rename at every invocation.

    ## Moving on ...
    """)


def get_ver(_allow_git_fetch=True):
    # Todo: There has to be a better library/method of doing this junk.
    from ksconf.vc.git import git_cmd
    ver_file = os.path.join("ksconf", "_version.py")
    # git_sha1 = git_cmd(["rev-parse", "HEAD"]).stdout[:12]
    vc_info = git_cmd(["show", "-s", "--abbrev=8", "--format=Git SHA1 %h committed on %cd",
                       "--date=format:%Y-%m-%d", "HEAD"]).stdout.strip()
    gitout = git_cmd(["describe", "--tags", "--always", "--dirty"])
    version = gitout.stdout.strip()
    version = version.lstrip("v")   # Tags format is v0.0.0
    del gitout

    # If version is hex string, assume there's an issue (aka running from pre-commit's install)
    # Pre-commit has it's own shallow clone that doesn't check out that tags we need to build the
    # package.  See https://github.com/pre-commit/pre-commit/issues/2610
    # Alternate pre-commit detection ideas: env PRE_COMMIT=1, or "pre-commit" in VIRTUAL_ENV or PATH
    if re.match(r'^[a-f0-9]+$', version):
        if _allow_git_fetch:
            # Fallback to valid, but silly version number; This seems better than failing
            print(f"Found unlikely version '{version}'.  Fetching git tags")
            git_cmd(["fetch", "--tags"])
            return get_ver(_allow_git_fetch=False)
        else:
            print("Can't determine version from git repository!")
            version = "0.0.0"

    # replace hash with local version format
    # XXX:  Wow. this needs unittests.  smh
    version = re.sub(r'-(\d+)-g([a-f0-9]+)((?:-dirty)?)', r'.dev\1+\2\3', version)
    version = re.sub(r'(\d+\.\d+)-dirty$', r'\1.dev0+dirty', version)
    version = version.replace("-dirty", ".dirty")
    print(f"Version:  {version}")
    build_no = 0
    ci_build_no = os.environ.get("GITHUB_RUN_NUMBER", None)
    if ci_build_no:
        build_no = 1000 + int(ci_build_no)
    code_block = dedent(f"""\
        # Version file autogenerated by the build process
        version = {version!r}
        build = {build_no!r}
        vcs_info = {vc_info!r}
        package_name = {package_name!r}

        if __name__ == '__main__':
            print('KSCONF_VERSION="{version}"\\nKSCONF_BUILD="{build_no}"\\nKSCONF_VCS_INFO="{vc_info}"')
        """)
    open(ver_file, "w").write(code_block)
    return version


DESCRIPTION = f"""\
# Ksconf Splunk CONFiguration tool

This utility handles common Splunk app maintenance tasks in an installable python package.
Specifically, this tool deals with many of the nuances of storing Splunk apps in a
version control system like git and pointing live Splunk apps to a working tree. Merging changes
from the live system's (local) folder to the version controlled (default) folder and dealing with
more than one layer of "default" are all supported tasks which are not native to Splunk.
Tasks like creating new Splunk apps from your local system while merging the 'local' folder into
'default' is also supported.

{package_rename_message}

Install with

    pip install {package_name}

Confirm installation with the following command:

    ksconf --version

To get an overview of all the CLI commands, run:

    ksconf --help

Help on specific command, such as 'unarchive', run:

    ksconf unarchive --help


Please see the [Official docs](https://ksconf.readthedocs.io/en/latest/) for more info.

"""


setup(name=package_name,
      version=get_ver(),
      description="KSCONF: Ksconf Splunk Configuration Tool",
      long_description=DESCRIPTION,
      long_description_content_type="text/markdown",
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Console",
          "Intended Audience :: System Administrators",
          "License :: OSI Approved :: Apache Software License",
          "Natural Language :: English",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX :: Linux",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Topic :: Utilities",
      ],
      python_requires='>=3.7',
      license="Apache Software License",
      keywords='ksconf splunk kinytre conf tool',
      author="Lowell Alleman",
      author_email="lowell@kintyre.co",
      url="https://github.com/Kintyre/ksconf",
      project_urls={
          "Documentation": "https://ksconf.readthedocs.io/",
          "Splunk app": "https://splunkbase.splunk.com/app/4383/",
      },
      packages=[
          "ksconf",
          "ksconf.app",
          "ksconf.builder",
          "ksconf.commands",
          "ksconf.conf",
          "ksconf.util",
          "ksconf.vc",
          "ksconf.ext",    # Third-party modules shipping with ksconf
      ],
      setup_requires=[
          "wheel",
      ],
      install_requires=[
          "entrypoints",
          "lxml",         # Added as a hard requirement to allow pre-commit to work out of the box
      ],
      # Wacky reason for this explained in ksconf/setup_entrypoints.py
      entry_points=get_entrypoints_setup(),
      # Not required, but useful.
      extras_require={
          "bash": ["argcomplete"],
          "thirdparty": [
              "splunk-sdk>=1.7.0"
          ],
      },
      include_package_data=True,
      zip_safe=True
      )



from __future__ import absolute_import, unicode_literals

import hashlib
import os
import re
import shutil
import tarfile
import tempfile
from typing import TextIO

from ksconf.app.manifest import AppManifest
from ksconf.combine import LayerCombiner
from ksconf.conf.merge import merge_app_local, merge_conf_dicts
from ksconf.conf.parser import conf_attr_boolean, parse_conf, update_conf
from ksconf.consts import KSCONF_DEBUG
from ksconf.vc.git import git_cmd


def find_conf_in_layers(app_dir, conf, *layers):
    if not layers:
        layers = ("local", "default")
    for layer in layers:
        conf_file = os.path.join(app_dir, layer, conf)
        if os.path.isfile(conf_file):
            return conf_file


def get_merged_conf(app_dir, conf, *layers):
    if not layers:
        # Last layer wins
        layers = ("default", "local")
    files = [os.path.join(app_dir, layer, conf) for layer in layers]
    confs = [parse_conf(path) for path in files if os.path.isfile(path)]
    return merge_conf_dicts(*confs)


def normalize_directory_mtime(path):
    """ Walk a tree and update the directory modification times to match the
    newest time of the children.  This results in a more predictable behavior
    over multiple executions.
    """
    for (root, dirs, files) in os.walk(path, topdown=False):
        nodes = dirs + files
        if not nodes:
            # Empty directories are somewhat unlikely.  We'll see
            continue
        mtime = max(os.stat(os.path.join(root, n)).st_mtime for n in nodes)
        os.utime(root, (mtime, mtime))


class PackagingException(Exception):
    pass


class AppPackager:

    def __init__(self, src_path, app_name: str, output: TextIO):
        self.src_path = src_path
        self.app_name = app_name
        self.build_dir = None
        self.app_dir = None
        self.output = output
        self._var_magic = None

    def cleanup(self):
        # Do we need  https://stackoverflow.com/a/21263493/315892  (Windows): -- See tests/cli_helper
        shutil.rmtree(self.build_dir)
        self.build_dir = None

    def expand_var(self, value):
        """ Expand a variable, if present

        :param str value:  String that main contain ``{{variable}}`` substitution.
        :return: Expanded value
        :rtype: str
        """
        return self._var_magic.expand(value)

    def expand_new_only(self, value):
        """ Expand a variable but return False if no substitution occurred

        :param str value:  String that may contain ``{{variable}}`` substitution.
        :return:  Expanded value if variables were expanded, else False
        :rtype: str
        """
        new_value = self._var_magic.expand(value)
        return new_value if new_value != value else False

    def combine(self, src, filters, layer_method="dir.d", allow_symlink=False):
        combiner = LayerCombiner(follow_symlink=allow_symlink, quiet=True)
        if layer_method == "dir.d":
            combiner.set_layer_root(src)
        elif layer_method == "disable":
            combiner.set_source_dirs([src])
        else:
            raise NotImplementedError(f"layer_method of '{layer_method}' is not supported.  "
                                      "Please use 'dir.d' or 'disable'.")
        for action, path in filters:
            combiner.add_layer_filter(action, path)
        combiner.combine(self.app_dir)
        self._var_magic.meta["layers"] = combiner.layer_names_used

    def blocklist(self, patterns):
        # XXX: Rewrite explicitly blocklist '.git' dir, because '.git*' wasn't working here. :=(

        # For now we just delete files out the build directory.  Not sophisticated, but it works
        # Do we need relwalker here?  relwalk
        from fnmatch import fnmatch
        for (root, dirs, files) in os.walk(self.build_dir, topdown=True):
            for fn in files:
                path = os.path.join(root, fn)
                for pattern in patterns:
                    if ("*" in pattern and fnmatch(path, pattern)) or fn == pattern:
                        self.output.write(f"Blocked file: {path}  (pattern: {pattern})\n")
                        os.unlink(path)
                        break
            for d in list(dirs):
                path = os.path.join(root, d)
                for pattern in patterns:
                    if ("*" in pattern and fnmatch(path, pattern)) or d == pattern:
                        self.output.write(f"Blocked dir:  {path}  (pattern: {pattern})\n")
                        dirs.remove(d)
                        shutil.rmtree(path)
                        break

    def merge_local(self):
        # XXX:  Rename this "promote_local()" ?
        """
        Find everything in local, if it has a corresponding file in default, merge.
        """
        # XXX: No logging/reporting done here :-(
        merge_app_local(self.app_dir)
        # Cleanup anything remaining in local
        self.block_local(report=False)

    def block_local(self, report=True):
        local_dir = os.path.join(self.app_dir, "local")
        if os.path.isdir(local_dir):
            if report:
                self.output.write("Removing local directory.\n")
            shutil.rmtree(local_dir)
        local_meta = os.path.join(self.app_dir, "metadata", "local.meta")
        if os.path.isfile(local_meta):
            if report:
                self.output.write("Removing local.meta\n")
            os.unlink(local_meta)

    def update_app_conf(self, version: str = None, build: str = None):
        """ Update version and/or build in ``apps.conf`` """
        app_settings = [
            ("launcher", "version", version),
            ("install", "build", build),
        ]
        appconf_file = find_conf_in_layers(self.app_dir, "app.conf") or \
            os.path.join(self.app_dir, "default", "app.conf")

        self.output.write(f"Updating app.conf file:  {appconf_file}\n")
        with update_conf(appconf_file, make_missing=True) as conf:
            for (stanza, attr, value) in app_settings:
                new_value = self.expand_new_only(value)
                if value:
                    if stanza not in conf:
                        conf[stanza] = {}
                    if new_value:
                        self.output.write(f"\tUpdate app.conf:  [{stanza}] "
                                          f"{attr} = {new_value} "
                                          f"(From {value})\n")
                        value = new_value
                    else:
                        self.output.write(f"\tUpdate app.conf:  [{stanza}] "
                                          f"{attr} = {value}\n")
                    conf[stanza][attr] = value

    def check(self):
        """ Run safety checks prior to building archive:

        1.  Set app name based on app.conf [package] id, if set.  Otherwise confirm that the package
            id and top-level folder names align.
        2.  Check for files or directories starting with ``.``, makes AppInspect very grumpy!
        """
        app_conf = get_merged_conf(self.app_dir, "app.conf")
        try:
            package_id = app_conf["package"]["id"]
            target_splunkbase = conf_attr_boolean(app_conf["package"]
                                                  .get("check_for_updates", "false"))
        except KeyError:
            self.output.write("Skipped folder and package id check due to missing app.conf entry\n")
            package_id = None

        if package_id:
            if not self.app_name or self.app_name == ".":
                self.output.write("Set app name from app.conf:  "
                                  f"{self.app_name}\n")
                self.app_name = package_id
            elif package_id != self.app_name:
                self.output.write(f"Top-level folder does not match the "
                                  f"package id:  folder: {self.app_name} "
                                  f"package id: {package_id}\n")
                if target_splunkbase:
                    raise PackagingException("Aborting build due to app name and package id "
                                             "discrepancy for public app")

        for root, dirs, files in os.walk(self.app_dir):
            for items, t in [(dirs, "directory"), (files, "file")]:
                for name in items:
                    if name[0] == ".":
                        self.output.write(f"Found hidden {t}:  {root}/{name}\n")

    def make_archive(self, filename):
        """ Create a compressed tarball of the build directory.
        """
        # type: (str) -> str
        # if os.path.isfile(filename):
        #    raise ValueError(f"Destination file already exists:  {filename}")
        app_name = self.expand_var(self.app_name)
        if app_name != self.app_name:
            self.output.write(f"Expanding template {self.app_name} to final "
                              f"app name: {app_name}\n")

        new_filename = self.expand_new_only(filename)
        if new_filename:
            self.output.write(f"Creating archive:  {new_filename}  (Expanded "
                              f"from '{os.path.basename(filename)}'\n")
            filename = new_filename
        else:
            self.output.write(f"Creating archive:  {filename}\n")

        normalize_directory_mtime(self.app_dir)
        with tarfile.open(filename, mode="w:gz") as spl:
            spl.add(self.app_dir, arcname=self.app_name)
        return filename

    def make_manifest(self, calculate_hash=True) -> AppManifest:
        """
        Create a manifest of the app's contents.
        """
        app_name = self.expand_var(self.app_name)
        manifest = AppManifest.from_filesystem(self.app_dir, name=app_name,
                                               calculate_hash=calculate_hash)
        manifest.source = self.src_path
        return manifest

    def __enter__(self):
        self.build_dir = tempfile.mkdtemp("-ksconf-package-build")
        if self.app_name == "." or "{{" in self.app_name:
            # Use a placeholder app name, otherwise build_dir == app_dir
            self.app_dir = os.path.join(self.build_dir, "app")
        else:
            self.app_dir = os.path.join(self.build_dir, self.app_name)
        self._var_magic = AppVarMagic(self.src_path, self.app_dir)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()


class AppVarMagicException(KeyError):
    pass


class AppVarMagic:
    """ A lazy loading dict-like object to fetch things like app version and such on demand. """

    def __init__(self, src_dir, build_dir, meta=None):
        self._cache = {}
        self.src_dir = src_dir
        self.build_dir = build_dir
        self.meta = meta or {}

    def expand(self, value):
        """ A simple Jinja2 like {{VAR}} substitution mechanism. """
        # type (str) -> str
        def replace(match_obj):
            var = match_obj.group(1)
            return self[var]
        if value:
            return re.sub(r"\{\{\s*([\w_]+)\s*\}\}", replace, value)
        return value

    def git_single_line(self, *args):
        out = git_cmd(args, cwd=self.src_dir)
        if out.returncode != 0:
            return f"git-errorcode-{out.returncode}"
        return out.stdout.strip()

    # START Variable fetching functions.  Be sure to add a docstring

    def get_version(self):
        """ Splunk app version fetched from app.conf """
        app_conf = get_merged_conf(self.build_dir, "app.conf")
        try:
            return app_conf["launcher"]["version"]
        except KeyError as e:
            raise AppVarMagicException(e)

    def get_build(self):
        """ Splunk app build fetched from app.conf """
        app_conf = get_merged_conf(self.build_dir, "app.conf")
        try:
            return app_conf["install"]["build"]
        except KeyError as e:
            raise AppVarMagicException(e)

    def get_app_id(self):
        """ Splunk app package id from app.conf """
        app_conf = get_merged_conf(self.build_dir, "app.conf")
        try:
            return app_conf["package"]["id"]
        except KeyError as e:
            raise AppVarMagicException(e)

    def get_git_tag(self):
        """ Git version tag using the ``git describe --tags`` command """
        tag = self.git_single_line("describe", "--tags", "--always", "--dirty")
        return re.sub(r'^(v|release|version)-?', "", tag)

    def get_git_last_rev(self):
        """ Git abbreviated rev of the last change of the app.  This may not be the same as HEAD. """
        return self.git_single_line("log", "-n1", "--pretty=format:%h", "--", ".")

    def get_git_head(self):
        """ Git HEAD rev abbreviated """
        return self.git_single_line("rev-parse", "--short", "HEAD")

    def get_layers_list(self):
        """ List of ksconf layers used. """
        layers = sorted(self.meta.get("layers"))
        if layers:
            return f'__{"__".join(layers)}__'
        else:
            return ""

    def get_layers_hash(self):
        """ Build a unique hash representing the combination of ksconf layers used. """
        DIGITS = 16
        layers_string = self["layers_list"]
        if layers_string:
            h = hashlib.sha256(layers_string.encode("utf-8"))
            return h.hexdigest()[:DIGITS]
        else:
            return "0" * DIGITS

    # END Variable fetching functions.

    def list_vars(self):
        """ Return a list of (variable, description) available in this class. """
        for name in dir(self):
            if name.startswith("get_"):
                var = name[4:]
                doc = getattr(self, name).__doc__.strip()
                yield (var, doc)

    def __getitem__(self, item):
        if item not in self._cache:
            self._cache[item] = self._get_expanded_var(item)
        return self._cache[item]

    def _get_expanded_var(self, item):
        get_funct_name = "get_" + item
        if hasattr(self, get_funct_name):
            try:
                funct = getattr(self, get_funct_name)
                return funct()
            except AppVarMagicException as e:
                if KSCONF_DEBUG in os.environ:
                    raise e
                return f"VAR-{item}-ERROR"
        else:
            raise KeyError(item)



"""
SUBCOMMAND:  ``ksconf merge --target=<CONF> <CONF> [ <CONF-n> ... ]``

Usage example:

.. code-block:: sh

    ksconf merge --target=master-props.conf /opt/splunk/etc/apps/*TA*/{default,local}/props.conf

"""
from __future__ import absolute_import, unicode_literals

import os

from ksconf.commands import ConfFileProxy, ConfFileType, KsconfCmd, dedent
from ksconf.conf.merge import merge_conf_files
from ksconf.conf.parser import PARSECONF_MID, PARSECONF_STRICT
from ksconf.consts import EXIT_CODE_SUCCESS
from ksconf.util.completers import conf_files_completer


class MergeCmd(KsconfCmd):
    help = "Merge two or more .conf files"
    description = dedent("""\
    Merge two or more .conf files into a single combined .conf file.
    This is similar to the way that Splunk logically combines the ``default`` and ``local``
    folders at runtime.
    """)
    maturity = "stable"

    def register_args(self, parser):
        parser.add_argument("conf", metavar="FILE", nargs="+",
                            help="The source configuration file(s) to collect settings from."
                            ).completer = conf_files_completer
        parser.add_argument("--target", "-t", metavar="FILE",
                            type=ConfFileType("r+", "none", parse_profile=PARSECONF_STRICT),
                            default=ConfFileProxy("<stdout>", "w", self.stdout), help=dedent("""\
            Save the merged configuration files to this target file.
            If not provided, the merged conf is written to standard output.""")
                            ).completer = conf_files_completer

        # This is helpful when writing bash expressions like MyApp/{default,local}/props.conf;
        # when either default or local may not be present.
        parser.add_argument("--ignore-missing", "-s", default=False, action="store_true",
                            help="Silently ignore any missing CONF files.")

        parser.add_argument("--dry-run", "-D", default=False, action="store_true", help=dedent("""\
            Enable dry-run mode.
            Instead of writing to TARGET, preview changes in 'diff' format.
            If TARGET doesn't exist, then show the merged file."""))
        parser.add_argument("--banner", "-b", default="", help=dedent("""\
            A banner or warning comment added to the top of the TARGET file.
            Used to discourage Splunk admins from editing an auto-generated file."""))

    def run(self, args):
        ''' Merge multiple configuration files into one '''
        self.parse_profile = PARSECONF_MID

        if args.ignore_missing:
            cfgs = [self.parse_conf(c) for c in args.conf if os.path.isfile(c) or c == "-"]
        else:
            cfgs = [self.parse_conf(conf) for conf in args.conf]

        merge_conf_files(args.target, cfgs, dry_run=args.dry_run,
                         banner_comment=args.banner)
        return EXIT_CODE_SUCCESS



import json
import os
import re
from io import BytesIO
from typing import TYPE_CHECKING, Any, List

from ksconf.util.file import ReluctantWriter

etree = None


if TYPE_CHECKING:
    from lxml.etree import ElementTree
else:
    ElementTree = Any


def _import_etree():
    # Lazy loaded etree; This prevents crashing at module import if not installed
    g = globals()
    from lxml import etree
    g["etree"] = etree


class FileReadlinesCache:
    """ Silly workaround for CDATA detection... """

    def __init__(self):
        self.cache = {}

    @staticmethod
    def convert_filename(filename):
        if filename.startswith("file:"):
            filename = filename[5:]
            if filename[0] == "/" and filename[2] == ":":
                # Example:   file:/c:/temp/.....
                filename = filename[1:]
            filename = os.path.normpath(filename)
        return filename

    def readlines(self, filename):
        if filename not in self.cache:
            self.cache[filename] = self._readlines(filename)
        return self.cache[filename]

    def _readlines(self, filename):
        filename = self.convert_filename(filename)
        with open(filename) as stream:
            return stream.readlines()


class SplunkSimpleXmlFormatter:
    keep_tags = {"latest", "earliest", "set", "label", "fieldset", "default", "search", "option"}

    def __init__(self):
        if etree is None:
            _import_etree()

    @classmethod
    def indent_tree(cls, elem: ElementTree, level=0, indent=2):
        # Copied from http://effbot.org/zone/element-lib.htm#prettyprint
        itxt = " " * indent
        i = "\n" + level * itxt
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + itxt
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                cls.indent_tree(elem, level + 1, indent)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    @classmethod
    def expand_tags(cls, elem: ElementTree, tags: set):
        """Keep <elem></elem> instead of shortening to <elem/>"""
        if elem.tag in tags and elem.text is None:
            # By setting this to an empty string (vs None), the trailing tag is kept
            elem.text = ""
        for c in elem:
            cls.expand_tags(c, tags)

    @staticmethod
    def cdata_tags(elem: ElementTree, tags: List[str]):
        """ Expand text to CDATA, if it isn't already. """
        cache = FileReadlinesCache()
        CDATA = "<![CDATA["

        def already_using_cdata(elem):
            # WARNING:  Super wonky hack!  I can't find any programmatic way to determine if .text
            # (or some part of it) contains CDATA.  There seems to be some hidden magic here,
            # in how this is implemented.  Patches welcome!

            # HACK optimization:   Cache the contents of the file to prevent dozens of re-reads
            lines = cache.readlines(elem.base)

            lineno = elem.sourceline - 1
            source_lines = lines[lineno:lineno + 2]
            for line in source_lines:
                if CDATA in line:
                    # print(f"Found source line:  {line}")
                    return True
            return False

        for tag in tags:
            for e in elem.findall(f".//{tag}"):
                if e.text and e.text.strip():
                    # Determine if the data is ALREADY in a CDATA element
                    # if isinstance(e.text, etree.CDATA):   # Doesn't work...
                    if already_using_cdata(e):
                        pass
                    elif re.search(r'[<>&]', e.text):
                        # print(f"SHOULD BE CDATA:   {e.text}")
                        # Convert text to CDATA
                        e.text = etree.CDATA(e.text)

    @staticmethod
    def guess_indent(elem: ElementTree, default=2):
        if elem.text:
            prefix = elem.text.strip("\r\n")
            indent = len(prefix) or default
            # print(f"Found indent={indent}")
        else:
            indent = default
        return indent

    @classmethod
    def format_json(cls, elem: ElementTree, indent=2):
        """  Format JSON data within a Dashboard Studio dashboard.   This is
        still pretty limited (for example, long searches still show up on a
        single line), but this give you at least a fighting change to figure
        out what's different.
        """

        # Abort if this isn't Dashboard Studio embedded in Simple XML
        if not elem.attrib.get("version") == "2":
            return

        tags = ["definition", "assets"]
        for tag in tags:
            for e in elem.findall(f".//{tag}"):
                if e.text:
                    data = json.loads(e.text)
                    output = json.dumps(data, indent=indent, sort_keys=False)
                    # Multiple lines?  Spread out CDATA header/trailer
                    if "\n" in output:
                        output = f"\n{output}\n"
                    e.text = etree.CDATA(output)

    @classmethod
    def format_xml(cls, src, dest, default_indent=2):
        # Copied from https://stackoverflow.com/a/5649263/315892
        parser = etree.XMLParser(resolve_entities=False, strip_cdata=False)
        document = etree.parse(src, parser)
        root = document.getroot()
        i = cls.guess_indent(root, default_indent)
        cls.indent_tree(root, indent=i)
        cls.cdata_tags(root, ["query"])
        cls.format_json(root, indent=i)
        cls.expand_tags(root, cls.keep_tags)

        b = BytesIO()
        document.write(b, pretty_print=True, encoding='utf-8')
        writer = ReluctantWriter(dest, "wb")
        with writer as f:
            f.write(b.getvalue().strip(b"\r\n"))
            f.write(b"\n")
        return writer.change_needed
